# -*- coding: utf-8 -*-
# This file is part of pypolsys, A python wrapper to `POLSYS_PLP`
# fortran90 package from Layne T. Watson, Steven M. Wise,
# Andrew J. Sommese, August, 1998.

# pypolsys is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pypolsys is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pypolsys.  If not, see <https://www.gnu.org/licenses/>.

r"""
pypolsys -- A python wrapper to `POLSYS_PLP` fortran90 homotopy solver
======================================================================
.. include::../README.md
    :start-line:4
    :raw:
"""

# On windows (gfortran+msvc) the fortran dll are put in a `.libs` folder
# by numpy.distutils and need to be in the path before import
# see https://pav.iki.fi/blog/2017-10-08/pywingfortran.html#building-python-wheels-with-fortran-for-windows

import os as _os
import sys as _sys
from platform import system as _system
if _system() == "Windows":
    extra_dll_dir = _os.path.join(_os.path.dirname(__file__), '.libs')
    # For python < 3.8 use path
    if _os.path.isdir(extra_dll_dir):
        _os.environ["PATH"] += _os.pathsep + extra_dll_dir
    # From python >= 3.8 ddl are managed with `os.add_dll_directory`
    if _sys.version_info >= (3, 8):
        if _os.path.isdir(extra_dll_dir):
            _os.add_dll_directory(extra_dll_dir)

# Not usefull to map all fortran modules
from .polsys import polsyswrap as polsys
from . import utils
from .utils import solve_univar
from .version import __version__

# Patch the docstring of the fortran module `polsyswrap` instead of the
# `polsys.xxx.so` file for pdoc3
# Rendering of f2py docstring is not perfect, but better than nothing.
__pdoc__ = {}
__pdoc__['polsys'] = polsys.__doc__



# -*- coding: utf-8 -*-

# This file is part of pypolsys, A python wrapper to `POLSYS_PLP`
# fortran90 package from Layne T. Watson, Steven M. Wise,
# Andrew J. Sommese, August, 1998.

# pypolsys is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pypolsys is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pypolsys.  If not, see <https://www.gnu.org/licenses/>.

import unittest
import doctest
import pypolsys
import numpy as np
from scipy.optimize import linear_sum_assignment
import os
import sys


def sort_ref_found(ref, r):
    """ Sort reference solution and the found solution using
    linear sum assignment.

    Parameters
    ----------
    ref : array
        The reference solution.
    r : array
        The found solution.

    Returns
    -------
    err : float
        The total error.
    """
    # Create Cost Matrix for linear sum assignment between
    # This function can also solve a generalization of the classic assignment problem
    # where the cost matrix is rectangular. If it has more rows than columns,
    # then not every row needs to be assigned to a column, and vice versa.
    # all combinations
    Nroots = r.shape[1]
    Nref = ref.shape[1]
    Cost = np.zeros((Nroots, Nref), dtype=float)
    for index, c in np.ndenumerate(Cost):
        Cost[index] = np.linalg.norm(ref[:, index[1]] - r[0:-1, index[0]])
    row_ind, col_ind = linear_sum_assignment(Cost)
    err = Cost[row_ind, col_ind].sum()
    return err

class TestBasic(unittest.TestCase):
    """ Test suite
    """

    def test_POLSYS_PLP1(self):
        """ Test example #1 of original `POLSYS_PLP` test suite.

        Two quadrics, no solutions at infinity, two real solutions.'
        It corresponds to 801/OUTPUT.DAT results
        """
        # Tolerance to compare with ref. solution
        tol = 1e-8
        # Define the monoms
        N = 2
        n_coef_per_eq = np.array([6, 6], dtype=np.int32)
        all_coef = np.array([-9.80e-04, 9.78e+05, -9.80e+00, -2.35e+02, 8.89e+04, -1.00e+00,
                             -1.00e-02, -9.84e-01, -2.97e+01, 9.87e-03, -1.24e-01, -2.50e-01],
                            dtype=complex)
        all_deg = np.zeros((np.sum(n_coef_per_eq), N), dtype=np.int32)
        all_deg[0, 0] = 2
        all_deg[1, 1] = 2
        all_deg[2, 0] = 1; all_deg[2, 1] = 1
        all_deg[3, 0] = 1
        all_deg[4, 1] = 1
        all_deg[5, 1] = 0
        all_deg[6, 0] = 2
        all_deg[7, 1] = 2
        all_deg[8, 0] = 1; all_deg[8, 1] = 1
        all_deg[9, 0] = 1
        all_deg[10, 1] = 1

        # Pass Polynomial coefficients and degree to POLSYS_PLP
        pypolsys.polsys.init_poly(N, n_coef_per_eq, all_coef, all_deg)
        # Create homogeneous partition
        part = pypolsys.utils.make_h_part(2)
        # Pass it to POLSYS_PLP
        pypolsys.polsys.init_partition(*part)
        # Solve with POLSYS_PLP
        bplp = pypolsys.polsys.solve(1e-8, 1e-14, 0.0)
        # Get back the roots
        r = pypolsys.polsys.myroots
        # Check with the reference solution
        x1_ref = np.array([9.08921229615392E-02-1.11115036326264E-17j,  2.34233851959134E+03-5.78409592525803E-12j, 1.61478579234411E-02+1.68496955498882E+00j, 1.61478579234120E-02-1.68496955498883E+00j])
        x2_ref = np.array([-9.11497098197499E-02+2.31749408304374E-17j, -7.88344824094162E-01+1.94092735010767E-15j, 2.67994739614473E-04+4.42802993973665E-03j, 2.67994739614405E-04-4.42802993973668E-03j])
        self.assertEqual(bplp, 4)
        # print(np.linalg.norm(np.sort(x1_ref) - np.sort(r[0, :])) )
        self.assertTrue(np.linalg.norm(np.sort(x1_ref)
                                       - np.sort(r[0, :])) < tol)
        self.assertTrue(np.linalg.norm(np.sort(x2_ref)
                                       - np.sort(r[1, :])) < tol)

    def test_3vars_homogeneous_partition(self):
        """ Consider the following example
        ```
        x**2 + y + z - 1 = 0
        x + y**2 + z - 1 = 0
        x + y + z**2 - 1 = 0
        ```
        > Cox, David, John Little, and Donal OShea. Ideals, varieties, and algorithms: an
        > introduction to computational algebraic geometry and commutative algebra.
        > Springer Science & Business Media, 2013, From page 122.

        With 5 solutions in C : (1, 0, 0) (0, 1, 0), (0, 0,1 ), (-1+√2, -1+√2, -1+√2),
        (-1 -√2, -1-√2, -1-√2).
        """
        # Tolerance to compare with ref. solution
        tol = 1e-8
        # 3 equations, 3 variables
        N = 3
        # Each containing 4 monoms
        n_coef_per_eq = np.array([4, 4, 4], dtype=np.int32)
        # The coefficients are provided as 1D array
        all_coef = np.array([1, 1,  1, -1,
                             1, 1,  1, -1,
                             1, 1,  1, -1], dtype=complex)
        # Degree of each monom
        all_deg = np.zeros((np.sum(n_coef_per_eq), N), dtype=np.int32)
        all_deg[0, 0] = 2
        all_deg[1, 1] = 1
        all_deg[2, 2] = 1
        all_deg[4, 0] = 1
        all_deg[5, 1] = 2
        all_deg[6, 2] = 1
        all_deg[8, 0] = 1
        all_deg[9, 1] = 1
        all_deg[10, 2] = 2
        # Pass it to POLSYS_PLP
        pypolsys.polsys.init_poly(N, n_coef_per_eq, all_coef, all_deg)
        # Create homogeneous partition
        part = pypolsys.utils.make_h_part(3)
        # Pass it to POLSYS_PLP
        pypolsys.polsys.init_partition(*part)
        # Found 8 solution, and track 8 paths
        bplp = pypolsys.polsys.solve(1e-8, 1e-14, 0.0)
        # Get the roots, array of size (N+1) x bplp
        r = pypolsys.polsys.myroots
        # Create References solution.
        ref = np.zeros((3, 5))
        sol = [(1., 0., 0.),
               (0., 1., 0.),
               (0., 0, 1.),
               (-1. + np.sqrt(2.), -1. + np.sqrt(2.), -1. + np.sqrt(2.)),
               (-1. - np.sqrt(2.), -1. - np.sqrt(2.), -1. - np.sqrt(2.))]
        for i, s in enumerate(sol):
            ref[:, i] = s

        # Need to sort both solutions in the same way
        err = sort_ref_found(ref, r)
        self.assertTrue(err < tol)

    def test_Cox_refine(self):
        """ Use 1-homogeneous partition and refine computation.
        """
        import sympy as sym
        # Define tols
        tol = 1e-4
        tracktol = 1e-8
        finaltol = 1e-8
        singtol = 0.

        tracktolr = 1e-14
        finaltolr = 1e-14
        singtol = 0.

        x, y, z = sym.symbols('x, y, z')
        pol = pypolsys.utils.fromSympy([sym.poly(x**2 + y + z - 1, (x, y, z)),
                                        sym.poly(x + y**2 + z - 1, (x, y, z)),
                                        sym.poly(x + y + z**2 - 1, (x, y, z))])
        # Pass it to POLSYS_PLP
        pypolsys.polsys.init_poly(*pol)
        # Create homogeneous partition
        part = pypolsys.utils.make_h_part(3)
        # Pass it to POLSYS_PLP
        pypolsys.polsys.init_partition(*part)
        # Solve with loose tolerances
        bplp = pypolsys.polsys.solve(tracktol, finaltol, finaltol)
        # Get the roots, array of size (N+1) x bplp
        r0 = pypolsys.polsys.myroots
        # Get status of the solving process
        pypolsys.polsys.report()
        # Create References solution.
        ref = np.zeros((3, 5))
        sol = [(1., 0., 0.),
               (0., 1., 0.),
               (0., 0, 1.),
               (-1. + np.sqrt(2.), -1. + np.sqrt(2.), -1. + np.sqrt(2.)),
               (-1. - np.sqrt(2.), -1. - np.sqrt(2.), -1. - np.sqrt(2.))]
        for i, s in enumerate(sol):
            ref[:, i] = s

        # Compute error with loose bounds
        err0 = sort_ref_found(ref, r0)
        # Refine one-half of the solutions
        path1 = np.array([1, 3, 5, 8])
        pypolsys.polsys.refine(path1, tracktolr, finaltolr, finaltol)
        path2 = np.array([2, 4, 6, 7])
        pypolsys.polsys.refine(path2, tracktolr, finaltolr, finaltol)
        # Get refine solution
        rr = pypolsys.polsys.myroots
        # Compute error with tight bounds
        errr = sort_ref_found(ref, rr)
        # Check error is lower than before
        self.assertTrue(errr*100 < err0)


    def test_univariate(self):
        """ Test univariate polynomial solve build from sympy.

        Consider the following example
        ```
        x**2 - 3*x + 2 = 0
        ```
        With 2 solutions in C : (2) (1)
        """
        # Tolerance to compare with ref. solution
        tol = 1e-8
        import sympy as sym
        x = sym.symbols('x')
        sparse = pypolsys.utils.fromSympy([sym.poly(x**2 - 3*x + 2, (x))])
        out4polsys = pypolsys.utils.toDense(*sparse)
        # Pass it to POLSYS_PLP
        pypolsys.polsys.init_poly(*out4polsys)
        # Create homogeneous partition
        part = pypolsys.utils.make_h_part(1)
        # Pass it to POLSYS_PLP
        pypolsys.polsys.init_partition(*part)
        # Solve
        bplp = pypolsys.polsys.solve(1e-8, 1e-14, 0.0, True)
        # Get the roots, array of size (N+1) x bplp
        r = pypolsys.polsys.myroots
        # Create References solution
        ref = np.array([1., 2.], dtype=complex)
        # Need to sort both solutions in the same way
        err = np.linalg.norm(ref - np.sort(r[0, :]))
        self.assertTrue(err < tol)
        self.assertEqual(bplp, 2)

    def test_horner_dense(self):
        """ Test horner, dense and standard approach on a simple test case,
        without solution at infnity.
        """

        import sympy as sym
        tol = 1e-4  # due to reference solution
        tracktol = 1e-8
        finaltol = 1e-14
        singtol = 0.
        # Example from https://nextjournal.com/saschatimme/homotopy-continuation-crash-course
        x, y = sym.symbols('x, y')
        monoms = pypolsys.utils.fromSympy([sym.poly(x**2 + 2*y, (x, y)),
                                           sym.poly(y**2 - 2, (x, y))])
        # 1-Homogeneous
        pypolsys.polsys.init_poly(*monoms)
        part = pypolsys.utils.make_h_part(2)
        pypolsys.polsys.init_partition(*part)
        # Solve and Store results
        b1 = pypolsys.polsys.solve(tracktol, finaltol, singtol)
        r1 = pypolsys.polsys.myroots.copy()

        # Use dense
        dmonoms = pypolsys.utils.toDense(*monoms)
        pypolsys.polsys.init_poly(*dmonoms)
        pypolsys.polsys.init_partition(*part)
        # Solve and Store results
        b1dense = pypolsys.polsys.solve(tracktol, finaltol, singtol, False)
        r1dense = pypolsys.polsys.myroots.copy()

        # Use dense AND Horner
        # Need to associate
        pypolsys.polsys.init_poly(*dmonoms)
        pypolsys.polsys.init_partition(*part)
        # Solve and Store results
        b1denseH = pypolsys.polsys.solve(tracktol, finaltol, singtol, True)
        r1denseH = pypolsys.polsys.myroots.copy()

        # Ref solution
        ref = np.array([[-2.51712e-32-1.68179j, 1.41421-3.69779e-32j],
                        [1.68179+1.54074e-33j, -1.41421-1.54074e-33j],
                        [2.51712e-32+1.68179j, 1.41421-3.69779e-32j],
                        [-1.68179-1.54074e-33j, -1.41421-1.54074e-33j]]).T
        # Need to sort both solutions in the same way
        err0 = sort_ref_found(ref, r1)        # standard
        err1 = sort_ref_found(ref, r1dense)   # dense
        err2 = sort_ref_found(ref, r1denseH)  # dense + horner
        self.assertTrue(err0 < tol)
        self.assertTrue(err1 < tol)
        self.assertTrue(err2 < tol)


class TestToyModel(unittest.TestCase):
    r""" More advanced Test case based on 3 dof toy model, with two parameters.

    The monomials and the reference solution are loaded from npz file.
        $$\begin{equation}        
        \left\lbrace
        \begin{aligned}
        - 2 \lambda - \mu - \nu + \left(\lambda + 2\right) \left(\lambda + \mu + 1\right) \left(\lambda + \nu + 1\right) - 2 &= 0, \\
        \left(\lambda + 2\right) \left(\lambda + \mu + 1\right) + \left(\lambda + 2\right) \left(\lambda + \nu + 1\right) + \left(\lambda + \mu + 1\right) \left(\lambda + \nu + 1\right) - 2 &= 0, \\
        6 \lambda + 2 \mu + 2 \nu + 8 &= 0.
        \end{aligned}
        \right.
        \end{equation}$$
    """

    @classmethod
    def setUpClass(cls):
        """ Load polynomial and reference solution from file.
        """
        # Load data
        input_file = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                    'examples/data/toy_model.npz'))
        data = np.load(input_file, allow_pickle=True)
        cls.out4polsys = data['out4polsys']
        cls.ref = data['ref'].T

    def test_1homogeneous(self):
        """ Use 1-homogeneous partition.
        """
        # Define tols
        tol = 1e-8
        tracktol = 1e-10
        finaltol = 1e-10
        singtol = 0.

        # 1-Homogeneous
        pypolsys.polsys.init_poly(*self.out4polsys)
        part = pypolsys.utils.make_h_part(3)
        pypolsys.polsys.init_partition(*part)
        # Solve and Store results
        b = pypolsys.polsys.solve(tracktol, finaltol, singtol)
        r = pypolsys.polsys.myroots
        # Need to sort both solutions in the same way
        err = sort_ref_found(self.ref, r)
        self.assertTrue(err < tol)

    def test_bezout(self):
        """ Use 1-homogeneous partition and compare with `bezout`.
        """
        # Define tols
        tol = 1e-8
        tracktol = 1e-10
        finaltol = 1e-10
        singtol = 0.

        # 1-Homogeneous
        pypolsys.polsys.init_poly(*self.out4polsys)
        part = pypolsys.utils.make_h_part(3)
        pypolsys.polsys.init_partition(*part)
        # Solve and compute Bezout
        b = pypolsys.polsys.solve(tracktol, finaltol, singtol)
        # Compute only Bezout
        B = pypolsys.polsys.bezout(singtol)
        # Check them
        self.assertEqual(b, B)

    def test_3_homogeneous(self):
        """ Use 3-homogeneous partition. Not optimal, but must find all solution.
        """
        # Define tols
        tol = 1e-8
        tracktol = 1e-10
        finaltol = 1e-10
        singtol = 0.

        # 1-Homogeneous
        pypolsys.polsys.init_poly(*self.out4polsys)
        part = pypolsys.utils.make_mh_part(3, [[1], [2], [3]])
        pypolsys.polsys.init_partition(*part)
        # Solve and Store results
        b = pypolsys.polsys.solve(tracktol, finaltol, singtol)
        r = pypolsys.polsys.myroots
        pypolsys.polsys.report()
        # Need to sort both solutions in the same way
        err = sort_ref_found(self.ref, r)
        self.assertTrue(err < tol)

    def test_plp_homogeneous(self):
        """ Use PLP partition. Not optimal, but must find all solution.
        """
        # Define tols
        tol = 1e-8
        tracktol = 1e-10
        finaltol = 1e-10
        singtol = 0.

        # PLP-partition
        pypolsys.polsys.init_poly(*self.out4polsys)
        part = pypolsys.utils.make_plp_part(3, [[[1], [2], [3]],
                                                [[1], [2, 3]],
                                                [[1, 2, 3]]])
        pypolsys.polsys.init_partition(*part)
        # Solve and Store results
        b = pypolsys.polsys.solve(tracktol, finaltol, singtol)
        r = pypolsys.polsys.myroots
        # Need to sort both solutions in the same way
        err = sort_ref_found(self.ref, r)
        self.assertTrue(err < tol)

    def test_3_homogeneousH(self):
        """ Use 3-homogeneous partition. Not optimal, but must find all solution.
        """
        # Define tols
        tol = 1e-8
        tracktol = 1e-10
        finaltol = 1e-10
        singtol = 0.

        # 1-Homogeneous
        dense = pypolsys.utils.toDense(*self.out4polsys)
        pypolsys.polsys.init_poly(*dense)
        part = pypolsys.utils.make_mh_part(3, [[1], [2], [3]])
        pypolsys.polsys.init_partition(*part)
        # Solve and Store results
        b = pypolsys.polsys.solve(tracktol, finaltol, singtol, True)
        r = pypolsys.polsys.myroots
        # Need to sort both solutions in the same way
        err = sort_ref_found(self.ref, r)
        self.assertTrue(err < tol)


if __name__ == '__main__':
    # Run unittest test suite
    print('> Running tests...')
    loadTestsFromTestCase = unittest.defaultTestLoader.loadTestsFromTestCase
    suite = loadTestsFromTestCase(TestBasic)
    suite.addTest(loadTestsFromTestCase(TestToyModel))
    suite.addTest(doctest.DocTestSuite(pypolsys.utils))
    # define the runner
    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(suite)
    # runner doesn't change exit status
    if result.wasSuccessful():
        sys.exit(0)
    else:
        sys.exit(1)



# -*- coding: utf-8 -*-

# This file is part of pypolsys, A python wrapper to `POLSYS_PLP`
# fortran90 package from Layne T. Watson, Steven M. Wise,
# Andrew J. Sommese, August, 1998.

# pypolsys is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pypolsys is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pypolsys.  If not, see <https://www.gnu.org/licenses/>.

"""
# pypolsys setup.

Define the fortran extension and its dependancies.
pypolsys can be either:
    - linked to lapack/blas (default on linux)
    - build with `POLSYS_PLP` lapack sources (default on windows)
The behavior can be specified with the install/developp flag `--buildLapack`

"""
import setuptools
import platform
import sys
import os
# Usefull to build f90 files
from numpy.distutils.core import Extension, setup
# To setup user options
from setuptools.command.install import install
from setuptools.command.develop import develop
# Use standard lib distutils for sdist (see numpy issues 7127)
from distutils.command.sdist import sdist


class UserOptionCommand(object):
    """ Mixin class to add user option support.

    based on https://stackoverflow.com/questions/18725137/how-to-obtain-arguments-passed-to-setup-py-from-pip-with-install-option
    """
    # Other options may be added here
    user_options = [
        ('buildLapack', None, 'Provide the required lapack/blas sources and build it.')]

    def initialize_options(self):
        """Initialize default options."""
        super().initialize_options()
        self.buildLapack = None

    def finalize_options(self):
        """Validate options."""
        # Nothing to do here
        super().finalize_options()

    def run(self):
        """Use options."""
        print('> Building with lapackBuild =', buildLapack)
        super().run()


# Update setuptool classes
class InstallCommand(UserOptionCommand, install):
    user_options = getattr(install, 'user_options', []) + UserOptionCommand.user_options


class DevelopCommand(UserOptionCommand, develop):
    user_options = getattr(develop, 'user_options', []) + UserOptionCommand.user_options


def ext_wrapper(buildLapack):
    """ Configure the fortran POLSYS_PLP extension.
    """
    # Base files
    sources = ['pypolsys/src/polsys.pyf',
               'pypolsys/801/polsys_plp.f90',
               'pypolsys/src/wrapper.f90']

    # Configure the POLSYS_PLP Extension building
    if buildLapack == 1:
        # Add lapack_plp to sources
        sources.append('pypolsys/801/lapack_plp.f')
        extra_link_args = []
    else:
        # Link to system lapack/blas
        extra_link_args = ['-llapack', '-lblas']
    return Extension(name='pypolsys.polsys',
                     sources=sources,
                     extra_compile_args=['-O3',
                                         '-ffast-math',
                                         '-funroll-loops'],  # "-fopenmp"
                     extra_link_args=extra_link_args)         # "-lgomp")


# Load README
with open("README.md", "r") as fh:
    long_description = fh.read()


# Load version
def version():
    """ Get version from version.py."""
    v = None
    with open(os.path.join('./pypolsys', 'version.py')) as f:
        for line in f:
            if line.lstrip().startswith('__version__'):
                v = line.split('=')[-1].strip().replace("'", "").replace('"', "")
                break
        return v


# Check platerform and options
buildLapack = None
if '--buildLapack' in sys.argv:
    buildLapack = 1
# On windows, by default, we used the lapack version shipped with POLSYS_PLP
this_os = platform.system().lower()
if this_os == 'windows':
    buildLapack = 1
print('> Building plateform =', this_os)
print('> Building with lapackBuild =', buildLapack)

# Fill setuptools
setup(
    name="pypolsys",
    version=version(),
    author="B. Nennig",
    author_email="benoit.nennig@isae-supmeca.fr",
    description=r"A python wrapper to `POLSYS_PLP` fortran90 package from Layne T. Watson,\
    Steven M. Wise, Andrew J. Sommese, August, 1998.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nennigb/pypolsys",
    # Use find_packages() to automatically discover all packages and subpackages
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'pypolsys': ['examples/data/*.npz']},
    # build f90 module
    ext_modules=[ext_wrapper(buildLapack)],
    install_requires=['numpy', 'scipy', 'sympy', 'setuptools'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent"],
    # tested with python 3.5 may works with previous py3 version...
    python_requires='>=3.5',
    # pass modified install and develop class to support user options
    cmdclass={
        'install': InstallCommand,
        'develop': DevelopCommand,
        'sdist': sdist,
        }
)



# -*- coding: utf-8 -*-

# This file is part of pypolsys, A python wrapper to `POLSYS_PLP`
# fortran90 package from Layne T. Watson, Steven M. Wise,
# Andrew J. Sommese, August, 1998.

# pypolsys is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pypolsys is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pypolsys.  If not, see <https://www.gnu.org/licenses/>.

"""
Define some utility functions to pass *easily* polynomials and variables
partitions to `POLSYS_PLP`.

Changing the partition has a strong impact on the number of paths to track
in the homotopy. `POLSYS_PLP` supports
  * 1-homogeneous, denoted by h,
  * m-homogeneous, denoted by mh,
  * partitioned linear product, denoted by plp.

For more details on `POLSYS_PLP` partition, See
> Wise, Steven M., Andrew J. Sommese, and Layne T. Watson. "Algorithm 801:
> POLSYS_PLP: A partitioned linear product homotopy code for solving polynomial
> systems of equations." ACM Transactions on Mathematical Software (TOMS)
> 26.1 (2000): 176-200.

For discussion on partitions and polynomials system types (dense, sparse, determinental), see
> Verschelde, Jan. "Polynomial homotopies for dense, sparse and determinantal
> systems", _arXiv preprint math/9907060_ (1999).
and the references therein.
"""

import numpy as np
from pypolsys import polsys


def fromSympy(P):
    """Instantiated polsys polynomial from sympy list of Poly.

    All variables should be present in the generator.

    Parameters
    ----------
    P : list
        A list of sympy poly objects.

    Returns
    -------
    N : int
        The number of variables and equations
    n_coef_per_eq : array
        The number of terms in each polynomial equation. Sum(n_coef_per_eq) give the
        total number of coefficeint
    all_coef : array
        Contains succesivelly all the coefficients of each monimial for each equation,
        starting by the first one.
    all_deg : array
        Contains succesivelly all the degrees of each monimial for each equation,
        starting by the first one.
        This is an array such the line number is the terms number and the column contain
        contains the degree of each variable.
        ex : if ther 4th term is (3.12+2j) x1^2 x2^3 x3^0, this yields
        all_deg(4, :) = [2, 3, 0] and all_coef(4) = 3.12+2j

    Remarks
    -------
    Use as a tuple, the output can be directly pass to `init_poly`.

    Examples
    --------
    >>> import sympy as sym
    >>> x, y = sym.symbols('x, y')
    >>> fromSympy([sym.poly(2*x + 3*y - x*y -3, (x,y)),\
                   sym.poly(3*x**2 + x*y - 1, (x,y))])  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    (2,
    array([4, 3]...),
    array([-3.+0.j,  3.+0.j,  2.+0.j, -1.+0.j,
           -1.+0.j,  1.+0.j,  3.+0.j]),
    array([[0, 0],
           [0, 1],
           [1, 0],
           [1, 1],
           [0, 0],
           [1, 1],
           [2, 0]]...))
    """
    # Number of variables
    N = len(P)
    # Check if all polys have the same Generator
    if all([p.gens == P[0].gens for p in P]) is False:
        raise ValueError('All polynomial should have the same generator (.gens).')

    coef_list = []
    deg_list = []

    # Conversion
    n_coef_per_eq = np.zeros((N,), dtype=np.int32)
    for n, p in enumerate(P):
        pd = p.as_dict()
        k = 0
        coef_list_n = []
        deg_list_n = []
        for key, val in pd.items():
            coef_list_n.append(val)
            deg_list_n.append(list(key))
            k += 1
        # store the number of terms for this polynomial
        n_coef_per_eq[n] = k
        # convert to array
        coef_arr_n = np.array(coef_list_n, dtype=complex)
        deg_arr_n = np.array(deg_list_n, dtype=np.int32)
        # Sort Them
        index = np.lexsort(np.fliplr(deg_arr_n).T)
        deg_list.append(deg_arr_n[index, :].copy())
        coef_list.append(coef_arr_n[index].copy())

    return N, n_coef_per_eq, np.hstack(coef_list), np.vstack(deg_list)


def from1Darray(P):
    """Instantiated polsys polynomial from 1D array corresponding to a univariate polynomial.

    The first term is the constant; assume dense representation.

    Parameters
    ----------
    P : iterable
        Coefficient of the polynomial ordered by ascending order.

    Returns
    -------
    N : int
        The number of variables and equations
    n_coef_per_eq : array
        The number of terms in each polynomial equation. Sum(n_coef_per_eq) give the
        total number of coefficeint
    all_coef : array
        Contains succesivelly all the coefficients of each monimial for each equation,
        starting by the first one.
    all_deg : array
        Contains succesivelly all the degrees of each monimial for each equation,
        starting by the first one.
        This is an array such the line number is the terms number and the column contain
        contains the degree of each variable.
        ex : if ther 4th term is (3.12+2j) x1^2 x2^3 x3^0, this yields
        all_deg(4, :) = [2, 3, 0] and all_coef(4) = 3.12+2j

    Remarks
    -------
    Use as a tuple, the output can be directly pass to `init_poly`.

    Examples
    --------
    Consider the following example
        ```
        x**2 - 3*x + 2 = 0
        ```
    With 2 solutions in C : {2, 1}
    >>> from1Darray([2., -3., 1.])  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    (1,
     array([3]...),
     array([ 2.+0.j, -3.+0.j,  1.+0.j]),
     array([[0],
            [1],
            [2]]...)
    """
    # Number of variables
    N = len(P)
    return (1, np.array([N], dtype=np.int32), np.array(P, dtype=complex),
            np.arange(0, N, dtype=np.int32).reshape(-1, 1))


def solve_univar(P, tracktol=1e-10, finaltol=1e-12, singtol=1e-14, dense=False):
    """Solve univariate polynomial ordered by ascending power order.

    This function is a short hand to solve univariate case where the calling
    sequence can be simplified.

    Parameters
    ----------
    P : iterable
        Coefficient of the polynomial ordered by ascending order.
        tracktol : float
        is the local error tolerance allowed the path tracker along
        the path.
    finaltol : float
        is the accuracy desired for the final solution.  It is used
        for both the absolute and relative errors in a mixed error criterion.
    singtol : float
        is the singularity test threshold used by `SINGSYS_PLP`.  If
        `singtol <= 0.0` on input, then `singtol` is reset to a default value.
    dense : bool
        if `True`, select the `TARGET_SYSTEM_USER` optimized for dense polynomial
        (horner). If `False` (default) the defaut `POLSYS_PLP TARGET_SYSTEM` is
        used. The default choice is safer.

    Returns
    -------
    roots : array
        The complex roots of the polynomial.

    Remarks
    --------
    For convenience the homogeneous variable due to the complex projective space
    are not return with this simplified interface.

    Examples
    --------
    Consider the following example
    ```
    x**2 - 3*x + 2 = 0
    ```
    With 2 solutions in C : {2, 1}
    >>> roots = solve_univar([2., -3., 1.])  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    >>> np.sort(roots.real)
    array([1., 2.])
    """
    out4polsys = from1Darray(P)
    # Pass it to POLSYS_PLP
    polsys.init_poly(*out4polsys)
    # Create homogeneous partition
    part = make_h_part(1)
    # Pass it to POLSYS_PLP
    polsys.init_partition(*part)
    # Solve
    bplp = polsys.solve(tracktol, finaltol, singtol, dense)
    # Get the roots, array of size (N+1) x bplp
    roots = polsys.myroots
    return roots[0, :]



def di(all_deg, n_coef_per_eq):
    """Return the degree of each equation and the total degree.

    This quantities are usefull to estimate the number of isolated solution
    of the polynomial system [Bezout Theorem].

    Parameters
    ----------
    all_deg : array
        Contains succesivelly all the degrees of each monimial for each equation,
        starting by the first one.

    n_coef_per_eq : array
        The number of terms in each polynomial equation. Sum(n_coef_per_eq) give the
        total number of coefficient

    Returns
    -------
    d : int
        The degree.
    dt : int
        the total degree.

    Examples
    --------
    >>> import sympy as sym
    >>> x, y = sym.symbols('x y')
    >>> N, n_coef_per_eq, all_coef, all_deg = fromSympy([sym.poly(2*x + 3*y - x*y -3, (x,y)),\
                                                         sym.poly(3*x**2 + x*y - 1, (x,y))])  # doctest: +NORMALIZE_WHITESPACE
    >>> di(all_deg, n_coef_per_eq)
    ([2, 2], 4)
    """
    d = []
    nini = 0
    for N in n_coef_per_eq:
        dmax = np.max(np.sum(all_deg[nini:nini+N, :], axis=1))
        d.append(dmax)
        nini += N

    dt = np.prod(d)

    return d, dt


def make_mh_part(N, var_list):
    """Create `POLSYS_PLP` arguments for m-homogeneous partition.

    Multi-homogeneous or m-homogeneous partition allow to group unknwons into sets,
    The partition will be the same for all equations.

    Parameters
    ----------
    N : int
        The number of variables and equations.
    var_list : list
        A list containing the group of variables used for all equations.
        All the variables must be present only once in each sublist.
        e. g. with 3 variables: [[1], [2], [3]], or [[1, 2], [3]] or
        [[3, 2], [1]].
        Note the index **must start at 1** to be compatible with fortran.

    Returns
    -------
    N : int
        The number of variable and equations.
    num_set :  array(N)
        The number of set for each polynomial equation.
    num_indices : array(N,  NUM_SETS)
        Contains the 'list' of all the coefficients, sorting such poly1 coef, poly2 coef...
    index : array (N, NUM_SETS, NUM_INDICES)
        Contains the 'list' of all the variable, in each set for each equation.

    Remarks
    -------
    Use as a tuple, the output can be directly pass to `init_partition`.

    Examples
    --------
    >>> make_mh_part(3, [[1, 2], [3]]) # doctest: +NORMALIZE_WHITESPACE
    (3,
    array([2., 2., 2.]),
    array([[2, 1],
           [2, 1],
           [2, 1]]),
    array([[[1., 2., 0.],
            [3., 0., 0.]],
          [[1., 2., 0.],
           [3., 0., 0.]],
          [[1., 2., 0.],
           [3., 0., 0.]]]))
    """
    # multi-homogeneous order
    m = len(var_list)
    # populated matrix for fortran
    num_set = np.ones((N,)) * m
    num_indice = np.array([len(part) for part in var_list])
    num_indices = np.repeat(num_indice.reshape(1, -1), N, axis=0)
    # create full index array
    # this the same for all row
    mindex = np.zeros((1, m, N))
    for i, part in enumerate(var_list):
        mindex[0, i, 0:num_indice[i]] = np.array(part)
    index = np.repeat(mindex, N, axis=0)
    return N, num_set, num_indices, index


def make_h_part(N):
    """Create `POLSYS_PLP` arguments for homogeneous partition.

    The partition will contain all variables, this equivalent to a 1-homogeneous
    partition. With this partition, the number of tracked paths will correspond to
    the total degree (Bézout Number).

    Parameters
    ----------
    N : int
        The number of variables and equations.

    Returns
    -------
    N : int
        The number of variable and equations.
    num_set :  array(N)
        The number of set for each polynomial equation.
    num_indices : array(N,  NUM_SETS)
        Contains the 'list' of all the coefficients, sorting such poly1 coef, poly2 coef...
    index : array (N, NUM_SETS, NUM_INDICES)
        Contains the 'list' of all the variable, in each set for each equation.

    Remarks
    -------
    Use as a tuple, the output can be directly pass to `init_partition`.
    """
    return make_mh_part(N, [list(range(1, N+1))])


def make_plp_part(N, var_list):
    """Create `POLSYS_PLP` arguments for plp-homogeneous partition.

    Compared to m-homogeneous partition, plp partition allows to change
    the partition for each equation. With this partition, the number of
    tracked paths will correspond to Bézout PLP number.

    Parameters
    ----------
    N : int
        The number of variables and equations.
    var_list : list
        A list containing the group of variables used for each equation.
        All the variables must be present only once in each equation sublist.
        e. g. with 3 variables, len(var_list) = 3 and the sublists looks like
        [[[1], [2, 3]], [[2], [1, 3]], [[3], [1, 2]]] or
        [[[1, 2, 3]], [[2], [1], [3]], [[3], [1, 2]]]
        Note the index **must start at 1** to be compatible with fortran.

    Returns
    -------
    N : int
        The number of variable and equations.
    num_set :  array(N)
        The number of set for each polynomial equation.
    num_indices : array(N,  NUM_SETS)
        Contains the 'list' of all the coefficients, sorting such poly1 coef, poly2 coef...
    index : array (N, NUM_SETS, NUM_INDICES)
        Contains the 'list' of all the variable, in each set for each equation.

    Remarks
    -------
    Use as a tuple, the output can be directly pass to `init_partition`.
    """
    # Check if N == len(var_list)
    if N != len(var_list):
        raise ValueError('len(var_list) is {} instead of {}'.format(len(var_list), N))

    # populated input matrices for fortran
    num_set = np.array([len(part) for part in var_list])
    num_indices = np.zeros((N, N))
    # create full index array
    # this the same for all row
    index = np.zeros((N, max(num_set), N))
    for i, part in enumerate(var_list):
        for j, var in enumerate(part):
            num_indices[i, j] = len(var)
            index[i, j, 0:len(var)] = np.array(var)

    return N, num_set, num_indices, index


def toDense(N, n_coef_per_eq, all_coef, all_deg, preserve=True):
    """Convert a *sparse* list of mononials suitable for `POLSYS_PLP`
    into a list of ordered monimials of a dense polynomials.

    Such preprocessing is **mandatory to use multiple variable Horner scheme**,
    ie `dense=True` in `polysys.solve`.
    For nearly dense polynomial, this method can significantly speed the polynomial
    and the Jacobian evaluation during the tracking.

      - Some absent monomial may be added to preserve the dense structure.
      - All monomials are listed left to right. The rightmost variable degree will
        change for each terms.
      - If a monimial is added to preserve the structure, its degree is set to 0,
        to avoid to change the Bezout Number computation in `POLSYS_PLP`.

    Parameters
    ----------
    N : int
        The number of variables and equations
    n_coef_per_eq : array
        The number of terms in each polynomial equation. Sum(n_coef_per_eq) give the
        total number of coefficeint
    all_coef : array
        Contains succesivelly all the coefficients of each monimial for each equation,
        starting by the first one.
    all_deg : array
        Contains succesivelly all the degrees of each monimial for each equation,
        starting by the first one.
        This is an array such the line number is the terms number and the column contain
        contains the degree of each variable.
        ex : if ther 4th term is (3.12+2j) x1^2 x2^3 x3^0, this yields
        all_deg(4, :) = [2, 3, 0] and all_coef(4) = 3.12+2j
    preserve : bool, optional
        Put the degree of vanishing coefficient to 0 to avoid to change Bezout number
        computation in `POLSYS_PLP` if dense structure is used. The default is True.

    Returns
    -------
    N : int
        The number of variables and equations
    n_coef_per_eq : array
        The number of terms in each polynomial equation. Sum(n_coef_per_eq) give the
        total number of coefficeint
    all_coef : array
        Contains succesivelly all the coefficients of each monimial for each equation,
        starting by the first one.
    all_deg : array
        Contains succesivelly all the degrees of each monimial for each equation,
        starting by the first one.
        This is an array such the line number is the terms number and the column contain
        contains the degree of each variable.

    Remarks
    -------
    Use as a tuple, the output can be directly pass to `init_poly`.
    Possible to enhance Horner speed if the highest degree is on the first variable.

    Examples
    --------
    >>> import sympy as sym
    >>> x, y = sym.symbols('x y')
    >>> sparse = fromSympy([sym.poly(3j*x**3 + 1, (x))])
    >>> toDense(*sparse)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
      > 200.0 % more coefs after conversion to dense.
    (1,
    array([4]...), array([1.+0.j, 0.+0.j, 0.+0.j, 0.+3.j]),
    array([[0],
          [0],
          [0],
         [3]]...))
    """
    # dmax contains the degree max in all variable for each equation
    dmax = np.zeros((N, N), dtype=np.int32)

    # setup new coefs matrix
    deg_list = []
    coef_list = []
    n_coef_per_eq_ = np.zeros_like(n_coef_per_eq)
    c = []
    # loop or equations
    start = 0
    for n in range(0, N):
        Nn = n_coef_per_eq[n]
        deg_n = all_deg[start:(start+Nn)]
        dmax[n, :] = np.max(deg_n, axis=0)
        # store ite
        deg_mat = np.zeros(dmax[n, :] + 1, dtype=complex)

        for coef, deg in zip(all_coef[start:(start+Nn)], deg_n):
            deg_mat[tuple(deg)] = coef
        c.append(deg_mat)
        # extract the indices and value
        deg_i = np.zeros((deg_mat.size, N), dtype=np.int32)
        coef_i = np.zeros((deg_mat.size,), dtype=complex)
        n_coef_per_eq_[n] = deg_mat.size
        k = 0
        for index, val in np.ndenumerate(deg_mat):
            if (val == 0) and preserve:
                # Put degree to 0 to avoid to change bezout Number for 0
                deg_i[k, :] = np.zeros_like(index)
            else:
                deg_i[k, :] = np.array(index)
            coef_i[k] = val
            k += 1
        deg_list.append(deg_i)
        coef_list.append(coef_i)

        start += Nn

    print('  > {} % more coefs after conversion to dense.'.format(np.sum(n_coef_per_eq_)
                                                         / np.sum(n_coef_per_eq)
                                                         * 100.))
    return N, n_coef_per_eq_, np.hstack(coef_list), np.vstack(deg_list)


# if __name__ == '__main__':
#     import doctest
#     doctest.testmod()
